==> bigloo3.1a.tar.gz: bigloo3.1a/api/multimedia/src/Llib/musicloop.scm <==
;*=====================================================================*/
;*    .../project/bigloo/api/multimedia/src/Llib/musicloop.scm         */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Fri May  2 09:58:46 2008                          */
;*    Last change :  Fri May 23 22:38:55 2008 (serrano)                */
;*    Copyright   :  2008 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    The implementation of the Music Event Loop                       */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __multimedia-music-event-loop

   (import __multimedia-music)
   
   (export (generic music-event-loop ::music . ::obj)
	   (generic music-event-loop-inner ::music ::obj ::obj ::obj ::obj)
	   (generic music-event-loop-handler ::music ::obj ::obj)
	   (generic music-event-loop-continue ::music ::obj ::obj ::obj ::obj)
	   (generic music-event-loop-reset! ::music)
	   (generic music-event-loop-abort! ::music)

	   (music-event-loop-parse-opt ::pair-nil)))

;*---------------------------------------------------------------------*/
;*    music-event-loop ::music ...                                     */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop o::music . obj)
   (with-access::music o (frequency %mutex %status %abort-loop %reset-loop)
      (mutex-lock! %mutex)
      (set! %abort-loop #f)
      (set! %reset-loop #f)
      (mutex-unlock! %mutex)
      (multiple-value-bind (onstate onmeta onerror onvol)
	 (music-event-loop-parse-opt obj)
	 ;; initial notfications
	 (music-update-status! o %status)
	 (when onstate (onstate %status))
	 (with-access::musicstatus %status (volume)
	    (when (and onvol (>fx volume 0)) (onvol volume)))
	 ;; enter the loop
	 (let loop ()
	    ;; the event loop is protected against timeout errors
	    (with-handler
	       (lambda (exc)
		  (music-event-loop-handler o exc onerror))
	       (music-event-loop-inner o onstate onmeta onerror onvol))
	    ;; loop back
	    (mutex-lock! %mutex)
	    (let ((abort %abort-loop))
	       (mutex-unlock! %mutex)
	       (unless (or %abort-loop (music-closed? o))
		  ;; flush the pending output
		  (when (music-event-loop-continue o onstate onmeta onerror onvol)
		     (sleep (* 2 frequency))
		     (loop))))))))

;*---------------------------------------------------------------------*/
;*    music-event-loop-handler ::music ...                             */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-handler o::music exc onerror)
   (tprint "MUSIC ERROR: " (find-runtime-type exc)
	   " (" (find-runtime-type o) ")")
   (exception-notify exc)
   (unless (or (&io-timeout-error? exc) (&io-write-error? exc))
      (let ((se (with-error-to-string (lambda () (exception-notify exc)))))
	 (when onerror (onerror se))
	 (raise exc))))
   
;*---------------------------------------------------------------------*/
;*    music-event-loop-inner ...                                       */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-inner m::music onstate onmeta onerror onvol)
   (with-access::music m (%mutex %abort-loop %reset-loop frequency %status)
      (mutex-lock! %mutex)
      (let loop ((stat1 (duplicate::musicstatus %status
			   (state 'init)))
		 (stat2 (instantiate::musicstatus
			   (state 'stop))))
	 (unless (or (music-closed? m) (music-%abort-loop m))
	    (mutex-unlock! %mutex)
	    (music-update-status! m stat2)
	    (with-access::musicstatus stat2 (state
					     err volume
					     song playlistid playlistlength)
;* 	       (tprint (musicstatus-state stat1) " / " state           */
;* 		       " " (eq? stat1 stat2))                          */
	       (when (and onstate
			  (or (not (eq? (musicstatus-state stat1) state))
			      (not (eq? (musicstatus-playlistid stat1) playlistid))
			      (not (eq? (musicstatus-song stat1) song))))
		  ;; onstate
		  (tprint "onstate: " stat2)
		  (onstate stat2)
		  (when (>fx playlistlength 0)
		     ;; onmeta
		     (let ((plist (music-playlist-get m)))
			(if (and (>=fx song 0) (<fx song playlistlength))
			    (onmeta (list-ref plist song) plist)
			    (onmeta #f plist)))))
	       (when (and onvol (not (eq? (musicstatus-volume stat1) volume)))
		  ;; onvolume
		  (onvol volume))
	       (when (and onerror err)
		  ;; onerror
		  (onerror err))
	       (sleep frequency)
	       ;; swap the two status
	       (mutex-lock! %mutex)
	       (when %reset-loop
		  (set! %reset-loop #f)
		  (musicstatus-state-set! stat2 'reset))
	       (loop stat2 stat1))))))

;*---------------------------------------------------------------------*/
;*    music-event-loop-continue ...                                    */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-continue o::music onstate onmeta onerror onvol)
   #unspecified)

;*---------------------------------------------------------------------*/
;*    music-event-loop-reset! ::music ...                              */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-reset! o::music)
   (with-access::music o (%mutex %reset-loop)
      (mutex-lock! %mutex)
      (set! %reset-loop #t)
      (mutex-unlock! %mutex)))

;*---------------------------------------------------------------------*/
;*    music-event-loop-abort! ...                                      */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-abort! o::music)
   (with-access::music o (%mutex %abort-loop)
      (mutex-lock! %mutex)
      (set! %abort-loop #t)
      (mutex-unlock! %mutex)))

;*---------------------------------------------------------------------*/
;*    music-event-loop-parse-opt ...                                   */
;*---------------------------------------------------------------------*/
(define (music-event-loop-parse-opt obj)
   
   (define (get-opt key arity)
      (let ((c (memq key obj)))
	 (when (and (pair? c) (pair? (cdr c)))
	    (cond
	       ((not (procedure? (cadr c)))
		(bigloo-type-error 'music-event-loop 'procedure (cadr c)))
	       ((not (correct-arity? (cadr c) arity))
		(error 'music-event-loop
		       (format "Bad ~a arity (expecting ~a)" key arity)
		       (cadr c)))
	       (else
		(cadr c))))))
   
   (values (get-opt :onstate 1)
	   (get-opt :onmeta 2)
	   (get-opt :onerror 1)
	   (get-opt :onvolume 1)))

==> bigloo3.2b.tar.gz: bigloo3.2b/api/multimedia/src/Llib/musicloop.scm <==
;*=====================================================================*/
;*    .../project/bigloo/api/multimedia/src/Llib/musicloop.scm         */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Fri May  2 09:58:46 2008                          */
;*    Last change :  Sat Jan  3 10:13:58 2009 (serrano)                */
;*    Copyright   :  2008-09 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    The implementation of the Music Event Loop                       */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __multimedia-music-event-loop

   (import __multimedia-music)
   
   (export (generic music-event-loop ::music . ::obj)
	   (generic music-event-loop-inner ::music ::obj ::obj ::obj ::obj)
	   (generic music-event-loop-reset! ::music)
	   (generic music-event-loop-abort! ::music)

	   (music-event-loop-parse-opt ::pair-nil)))

;*---------------------------------------------------------------------*/
;*    music-event-loop ::music ...                                     */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop o::music . obj)
   (with-access::music o (frequency %mutex %loop-mutex %loop-condv
				    %status %abort-loop %reset-loop)
      (mutex-lock! %loop-mutex)
      (set! %abort-loop #f)
      (set! %reset-loop #f)
      (mutex-unlock! %loop-mutex)
      (multiple-value-bind (onstate onmeta onerror onvol)
	 (music-event-loop-parse-opt obj)
	 ;; setup state
	 (with-access::musicstatus %status (state volume)
	    (set! volume 0)
	    (set! state 'init))
	 ;; enter the loop
	 (unwind-protect
	    ;; the event loop is protected against timeout errors
	    (music-event-loop-inner o onstate onmeta onerror onvol)
	    ;; signal that the loop is done
	    (begin
	       (mutex-lock! %loop-mutex)
	       (condition-variable-broadcast! %loop-condv)
	       (mutex-unlock! %loop-mutex))))))

;*---------------------------------------------------------------------*/
;*    music-event-loop-inner ...                                       */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-inner m::music onstate onmeta onerror onvol)
   
   (define (newstate? stat2 stat1)
      (with-access::musicstatus stat2 (state song playlistid)
	 (or (not (eq? (musicstatus-state stat1) state))
	     (not (eq? (musicstatus-playlistid stat1) playlistid))
	     (not (eq? (musicstatus-song stat1) song)))))
   
   (define (newvolume? stat2 stat1)
      (with-access::musicstatus stat2 (volume)
	 (not (eq? (musicstatus-volume stat1) volume))))
   
   (with-access::music m (%loop-mutex %abort-loop %reset-loop frequency %status)
      (mutex-lock! %loop-mutex)
      (let loop ((stat1 (duplicate::musicstatus %status
			   (state 'init)))
		 (stat2 (instantiate::musicstatus
			   (state 'stop))))
	 (let ((stop (or (music-closed? m) (music-%abort-loop m))))
	    (mutex-unlock! %loop-mutex)
	    (unless stop
	       (music-update-status! m stat2)

	       (when (newstate? stat2 stat1)
		  ;; onstate
		  (when onstate (onstate stat2))
		  
		  ;; onmeta
		  (with-access::musicstatus stat2 (song playlistlength)
		     (when (>fx playlistlength 0)
			(let ((plist (music-playlist-get m)))
			   (if (and (>=fx song 0) (<fx song (length plist)))
			       (onmeta (list-ref plist song) plist)
			       (onmeta #f plist))))))
	       
	       (when (and onvol (newvolume? stat2 stat1))
		  ;; onvolume
		  (onvol (musicstatus-volume stat2)))
	       
	       (when (and onerror (musicstatus-err stat2))
		  ;; onerror
		  (onerror (musicstatus-err stat2)))

	       ;; wait a little bit
	       (sleep frequency)
	       
	       ;; swap the two status
	       (mutex-lock! %loop-mutex)
	       (when %reset-loop
		  (set! %reset-loop #f)
		  (musicstatus-state-set! stat2 'reset))

	       ;; loop back
	       (loop stat2 stat1))))))

;*---------------------------------------------------------------------*/
;*    music-event-loop-reset! ::music ...                              */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-reset! o::music)
   (with-access::music o (%loop-mutex %reset-loop)
      (mutex-lock! %loop-mutex)
      (set! %reset-loop #t)
      (mutex-unlock! %loop-mutex)))

;*---------------------------------------------------------------------*/
;*    music-event-loop-abort! ...                                      */
;*---------------------------------------------------------------------*/
(define-generic (music-event-loop-abort! o::music)
   (with-access::music o (%loop-mutex %abort-loop %loop-condv)
      (mutex-lock! %loop-mutex)
      (set! %abort-loop #t)
      (condition-variable-wait! %loop-condv %loop-mutex)
      (mutex-unlock! %loop-mutex)))

;*---------------------------------------------------------------------*/
;*    music-event-loop-parse-opt ...                                   */
;*---------------------------------------------------------------------*/
(define (music-event-loop-parse-opt obj)
   
   (define (get-opt key arity)
      (let ((c (memq key obj)))
	 (when (and (pair? c) (pair? (cdr c)))
	    (cond
	       ((not (procedure? (cadr c)))
		(bigloo-type-error 'music-event-loop 'procedure (cadr c)))
	       ((not (correct-arity? (cadr c) arity))
		(error 'music-event-loop
		       (format "Bad ~a arity (expecting ~a)" key arity)
		       (cadr c)))
	       (else
		(cadr c))))))
   
   (values (get-opt :onstate 1)
	   (get-opt :onmeta 2)
	   (get-opt :onerror 1)
	   (get-opt :onvolume 1)))
